from typing import List, Optional

from benchling_api_client.v2.stable.api.label_templates import list_label_templates

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import LabelTemplate
from benchling_sdk.services.v2.base_service import BaseService


class LabelTemplateService(BaseService):
    """
    Label Templates.

    List label templates.

    See https://benchling.com/api/reference#/Label%20Templates
    """

    @api_method
    def get_list(self, registry_id: str, name: Optional[str] = None) -> List[LabelTemplate]:
        """
        List label templates.

        See https://benchling.com/api/reference#/Label%20Templates/listLabelTemplates

        :param registry_id: The ID of the registry for which to list label templates
        :param name: The name of the label template
        :return: A list of label templates
        :rtype: List[LabelTemplate]
        """
        response = list_label_templates.sync_detailed(
            client=self.client, registry_id=registry_id, name=none_as_unset(name)
        )
        results = model_from_detailed(response)
        return results.label_templates



from typing import Iterable, List, Optional

from benchling_api_client.v2.stable.api.requests import (
    bulk_create_request_tasks,
    bulk_get_requests,
    bulk_update_request_tasks,
    create_request,
    execute_requests_sample_groups,
    get_request,
    get_request_fulfillment,
    get_request_response,
    list_request_fulfillments,
    list_requests,
    patch_request,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    Request,
    RequestCreate,
    RequestFulfillment,
    RequestFulfillmentsPaginatedList,
    RequestResponse,
    RequestsPaginatedList,
    RequestStatus,
    RequestTaskBase,
    RequestTasksBulkCreate,
    RequestTasksBulkCreateRequest,
    RequestTasksBulkCreateResponse,
    RequestTasksBulkUpdateRequest,
    RequestTasksBulkUpdateResponse,
    RequestUpdate,
    SampleGroupsStatusUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class RequestService(BaseService):
    """
    Requests.

    Requests allow scientists and teams to collaborate around experimental assays and workflows.

    See https://benchling.com/api/reference#/Requests
    """

    @api_method
    def get_by_id(self, request_id: str) -> Request:
        """
        Get a Request by ID.

        See https://benchling.com/api/reference#/Requests/getRequest
        """
        response = get_request.sync_detailed(client=self.client, request_id=request_id)
        return model_from_detailed(response)

    @api_method
    def _requests_page(
        self,
        schema_id: str,
        request_status: Optional[RequestStatus] = None,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        next_token: NextToken = None,
        page_size: Optional[int] = None,
    ) -> Response[RequestsPaginatedList]:
        response = list_requests.sync_detailed(
            client=self.client,
            schema_id=schema_id,
            request_status=none_as_unset(request_status),
            min_created_time=none_as_unset(min_created_time),
            max_created_time=none_as_unset(max_created_time),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response

    def list(
        self,
        schema_id: str,
        request_status: Optional[RequestStatus] = None,
        min_created_time: Optional[int] = None,
        max_created_time: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Request]:
        """
        List Requests.

        See https://benchling.com/api/reference#/Requests/listRequests
        """

        def api_call(next_token: NextToken) -> Response[RequestsPaginatedList]:
            return self._requests_page(
                schema_id=schema_id,
                request_status=request_status,
                min_created_time=min_created_time,
                max_created_time=max_created_time,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: RequestsPaginatedList) -> Optional[List[Request]]:
            return body.requests

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(
        self, *, request_ids: Optional[Iterable[str]] = None, display_ids: Optional[Iterable[str]] = None
    ) -> Optional[List[Request]]:
        """
        Bulk get Requests.

        See https://benchling.com/api/reference#/Requests/bulkGetRequests
        """
        request_id_string = optional_array_query_param(request_ids)
        display_id_string = optional_array_query_param(display_ids)
        response = bulk_get_requests.sync_detailed(
            client=self.client,
            request_ids=none_as_unset(request_id_string),
            display_ids=none_as_unset(display_id_string),
        )
        requests_list = model_from_detailed(response)
        return requests_list.requests

    @api_method
    def request_response(self, request_id: str) -> RequestResponse:
        """
        Get a Request's response.

        See https://benchling.com/api/reference#/Requests/getRequestResponse
        """
        response = get_request_response.sync_detailed(client=self.client, request_id=request_id)
        return model_from_detailed(response)

    @api_method
    def create(self, request: RequestCreate) -> Request:
        """
        Create a Request.

        See https://benchling.com/api/reference#/Requests/createRequest
        """
        response = create_request.sync_detailed(client=self.client, json_body=request)
        return model_from_detailed(response)

    @api_method
    def update(self, request_id: str, request: RequestUpdate) -> Request:
        """
        Update a Request.

        See https://benchling.com/api/reference#/Requests/patchRequest
        """
        response = patch_request.sync_detailed(client=self.client, request_id=request_id, json_body=request)
        return model_from_detailed(response)

    @api_method
    def execute_sample_groups(self, request_id: str, sample_groups: SampleGroupsStatusUpdate) -> None:
        """
        Update the status of sample groups in a Request.

        See https://benchling.com/api/reference#/Requests/executeRequestsSampleGroups
        """
        response = execute_requests_sample_groups.sync_detailed(
            client=self.client, request_id=request_id, json_body=sample_groups
        )
        return model_from_detailed(response)

    @api_method
    def request_fulfillment(self, request_fulfillment_id: str) -> RequestFulfillment:
        """
        Get a Request's fulfillment.

        See https://benchling.com/api/reference#/Requests/getRequestFulfillment
        """
        response = get_request_fulfillment.sync_detailed(
            client=self.client, request_fulfillment_id=request_fulfillment_id
        )
        return model_from_detailed(response)

    @api_method
    def _entry_request_fulfillments_page(
        self,
        entry_id: str,
        modified_at: Optional[str] = None,
        next_token: NextToken = None,
        page_size: Optional[int] = None,
    ) -> Response[RequestFulfillmentsPaginatedList]:
        response = list_request_fulfillments.sync_detailed(
            client=self.client,
            entry_id=entry_id,
            modified_at=none_as_unset(modified_at),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def entry_request_fulfillments(
        self, entry_id: str, modified_at: Optional[str] = None, page_size: Optional[int] = None
    ) -> PageIterator[RequestFulfillment]:
        """
        List Request Fulfillments.

        See https://benchling.com/api/reference#/Requests/listRequestFulfillments
        """

        def api_call(next_token: NextToken) -> Response[RequestFulfillmentsPaginatedList]:
            return self._entry_request_fulfillments_page(
                entry_id=entry_id,
                modified_at=modified_at,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: RequestFulfillmentsPaginatedList) -> Optional[List[RequestFulfillment]]:
            return body.request_fulfillments

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_create_tasks(
        self, request_id: str, tasks: Iterable[RequestTasksBulkCreate]
    ) -> RequestTasksBulkCreateResponse:
        """
        Create tasks for a Request.

        See https://benchling.com/api/reference#/Requests/bulkCreateRequestTasks
        """
        create = RequestTasksBulkCreateRequest(tasks=list(tasks))
        response = bulk_create_request_tasks.sync_detailed(
            client=self.client, request_id=request_id, json_body=create
        )
        return model_from_detailed(response)

    @api_method
    def bulk_update_tasks(
        self, request_id: str, tasks: Iterable[RequestTaskBase]
    ) -> RequestTasksBulkUpdateResponse:
        """
        Update tasks for a Request.

        See https://benchling.com/api/reference#/Requests/bulkUpdateRequestTasks
        """
        update = RequestTasksBulkUpdateRequest(tasks=list(tasks))
        response = bulk_update_request_tasks.sync_detailed(
            client=self.client, request_id=request_id, json_body=update
        )
        return model_from_detailed(response)



from typing import List, Optional

from benchling_api_client.v2.stable.api.printers import list_printers

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.models import Printer
from benchling_sdk.services.v2.base_service import BaseService


class PrinterService(BaseService):
    """
    Printers.

    List printers.

    See https://benchling.com/api/reference#/Printers
    """

    @api_method
    def get_list(self, registry_id: str, name: Optional[str] = None) -> List[Printer]:
        """
        List printers.

        See https://benchling.com/api/reference#/Printers/listPrinters

        :param registry_id: ID of the registry to list printers from.
        :param name: The name of the printer.
        :return: A list of printers
        :rtype: List[Printer]
        """
        response = list_printers.sync_detailed(
            client=self.client, registry_id=registry_id, name=none_as_unset(name)
        )
        results = model_from_detailed(response)
        return results.label_printers



from abc import ABC, abstractmethod


class BaseDecryptionProvider(ABC):
    """
    Provides a way to decrypt encrypted messages.

    Various implementations might use AWS KMS, Azure, etc.
    """

    @abstractmethod
    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypt secrets.

        Receives encrypted cipher text and returns the decrypted plain text.
        """
        pass



from dataclasses import dataclass
from http import HTTPStatus
import logging
from typing import Optional, Tuple, Union

import backoff
from typing_extensions import Literal

from benchling_sdk.errors import BenchlingError

# By default, silence all except critical logs from `backoff`.
# Otherwise, `backoff` logs an error when giving up, but this is not necessary
# because it already re-raises the `BenchlingError`.
logging.getLogger("backoff").setLevel(logging.CRITICAL)


@dataclass(frozen=True, eq=True)
class RetryStrategy(object):
    """Specify a strategy for automatically retrying failed HTTP calls."""

    # Passing in None results in unbounded retries
    max_tries: Optional[int] = 5
    # Wait time between calls is backoff_factor * 2^n, where n starts at 0
    backoff_factor: float = 0.5
    status_codes_to_retry: Tuple[HTTPStatus, ...] = (
        HTTPStatus.TOO_MANY_REQUESTS,
        HTTPStatus.BAD_GATEWAY,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    )

    @staticmethod
    def no_retries() -> "RetryStrategy":
        """Construct a noop retry strategy that will not retry."""
        return RetryStrategy(0, 0, ())


_sentinel = object()


def retry_method(f):
    """
    Decorate a function to retry wrapped method if a BenchlingError is raised.

    The wrapped method must be a method on a subclass of `BaseService`,
    because the default retry strategy is taken from the `self` argument of the method.

    The decorator also adds an additional kwarg `retry_strategy` to the method signature.
    This can be used to override the service's default retry strategy.
    """
    # Inline import to avoid circular dependency
    from benchling_sdk.services.v2.base_service import BaseService

    def func_with_retry_strategy_arg(
        self: BaseService,
        *args,
        retry_strategy: Union[RetryStrategy, Literal[_sentinel]] = _sentinel,
        **kwargs
    ):
        if retry_strategy is _sentinel:
            retry_strategy = self.retry_strategy

        def should_retry(e: BenchlingError) -> bool:
            if retry_strategy.status_codes_to_retry is None:
                return True
            return e.status_code in retry_strategy.status_codes_to_retry

        @backoff.on_exception(
            backoff.expo,
            BenchlingError,
            max_tries=retry_strategy.max_tries,
            giveup=lambda e: not should_retry(e),
            factor=retry_strategy.backoff_factor,
        )
        def func_with_retries(*fargs, **fkwargs):
            return f(*fargs, **fkwargs)

        return func_with_retries(self, *args, **kwargs)

    return func_with_retry_strategy_arg



from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.client_helpers import v2_alpha_client
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.services.v2.alpha.v2_alpha_app_service import V2AlphaAppService
from benchling_sdk.services.v2.base_service import BaseService


class V2AlphaService(BaseService):
    """
    V2-alpha.

    Alpha endpoints have different stability guidelines than other stable endpoints.

    See https://benchling.com/api/v2-alpha/reference
    """

    _app_service: V2AlphaAppService

    def __init__(self, client: Client, retry_strategy: RetryStrategy = RetryStrategy()):
        """
        Initialize a v2-alpha service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)
        alpha_client = v2_alpha_client(self.client)
        self._app_service = V2AlphaAppService(alpha_client, retry_strategy)

    @property
    def apps(self) -> V2AlphaAppService:
        """
        V2-Alpha Apps.

        Create and manage Apps on your tenant.

        https://benchling.com/api/v2-alpha/reference?stability=not-available#/Apps
        """
        return self._app_service
